/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is MoreLayoutsforThunderbird.
 *
 * The Initial Developer of the Original Code is
 *   alta88 <alta88@gmail.com>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("resource:///modules/gloda/log4moz.js");

var MoreLayouts = {
  Cc: Components.classes,
  Ci: Components.interfaces,

  get log() {
    let logger = Log4Moz.getConfiguredLogger("MoreLayouts");
    this.__defineGetter__("log", function() { return logger });
    return this.log;
  },

  get paneConfig() {
    let pref = "mail.pane_config.dynamic";
    return gPrefBranch.getIntPref(pref);
  },

  set paneConfig(val) {
    let pref = "mail.pane_config.dynamic";
    gPrefBranch.setIntPref(pref, val);
  },

  get fullscreenPaneConfig() {
    let pref = "extensions.morelayoutsforthunderbird.fullscreenPaneConfig";
    return gPrefBranch.getIntPref(pref);
  },

  gCurrentPaneConfig:              kStandardPaneConfig,
  gAccountCentralLoaded:           true,
  // kStandardPaneConfig = 0,
  // kWidePaneConfig     = 1,
  // kVerticalPaneConfig = 2,
  kWideThreadPaneConfig:           3,
  kStackedPaneConfig:              4,
  kAccountCentral:                 5,

  dynamicViewName:                 ["standard",
                                    "widemessage",
                                    "vertical",
                                    "widethread",
                                    "stacked",
                                    "accountcentral"],

  dynamicViewResetDOMType:         [3, 4],

  dynamicIdsMessagePane:           ["messagesBox",
                                    "mailContent",
                                    "threadPaneBox",
                                    "messagesBox",
                                    "messagesBox"],

  dynamicIdsThreadPane:            ["messagesBox",
                                    "messagesBox",
                                    "messagesBox",
                                    "mailContent",
                                    "folderPaneBox"],

  dynamicIdsThreadPaneSplitter:    ["messagesBox",
                                    "mailContent",
                                    "threadPaneBox",
                                    "mailContent",
                                    "folderPaneBox"],

  // For Orient states, 0=horizontal, 1=vertical.
  dynamicThreadPaneSplitterOrient: [1, 1, 0, 1, 1],
  dynamicMessagePaneBoxFlex:       [2, 0, 2, 2, 2],
  dynamicThreadPaneBoxFlex:        [1, 1, 1, 1, 0],

  // For F8 toggle.
  get messagePaneToggleMode() {
    let pref = "extensions.morelayoutsforthunderbird.messagePaneToggleMode";
    let mode = gPrefBranch.getIntPref(pref);
    let valid = [this.kToggleNormalMin,
                 this.kToggleNormalMax,
                 this.kToggleCycle].indexOf(mode) != -1 ? true : false;
    return valid ? mode : this.kToggleNormalMin;
  },

  gCurrentMessagePaneConfig:       null,
  gPriorMessagePaneConfig:         null,
  kMessagePaneNormal:              "normal",
  kMessagePaneMin:                 "min",
  kMessagePaneMax:                 "max",
  kToggleNormalMin:                0,
  kToggleNormalMax:                1,
  kToggleCycle:                    2,

  e: function(element) {
    return document.getElementById(element);
  },

  /**
   * Restore msgWindow object.  This fixes Bug 492216 and Bug 636874.
   * They aren't really bugs though, just misplaced expectations based on
   * overly optimistic doc.
   */
  rerootMsgWindow: function() {
this.log.debug("rerootMsgWindow: START ");
    // Global object.
    msgWindow = this.Cc["@mozilla.org/messenger/msgwindow;1"].
                createInstance(this.Ci.nsIMsgWindow);
    msgWindow.notificationCallbacks = new BadCertHandler();
    InitMsgWindow();
    messenger.setWindow(window, msgWindow);

    // Tab objects.
    let tabmail = MoreLayouts.e("tabmail");
    for each (let [index, tab] in Iterator(tabmail.tabInfo)) {
      if (tab.mode.type == "folder")
        tab.folderDisplay.messenger.setWindow(window, msgWindow);
      if (tab.mode.type == "message")
        tab.messageDisplay.folderDisplay.messenger.setWindow(window, msgWindow);
    }

this.log.debug("rerootMsgWindow: completed ");
  },

  /**
   * Restore threadpane references and view.
   */
  rerootThreadPane: function() {
this.log.debug("rerootThreadPane: START ");
    if (!gDBView)
      return;

    let tabmail = MoreLayouts.e("tabmail");
    let threadTree = GetThreadTree();
    let treeView = gFolderDisplay.view.dbView;

    for each (let [index, tab] in Iterator(tabmail.tabInfo)) {
      if (tab.mode.type == "folder") {
        tab.folderDisplay.tree = threadTree;
        tab.folderDisplay.treeBox = tab.folderDisplay.tree.boxObject.
                                    QueryInterface(this.Ci.nsITreeBoxObject);
this.log.debug("rerootThreadPane: folder tab - "+tab.title);
      }
    }

    if (treeView) {
      threadTree.boxObject.view = treeView;
this.log.debug("rerootThreadPane: treeView completed");
    }
this.log.debug("rerootThreadPane: done ");
  },

  /**
   * New FolderDisplayWidget._showAccountCentral().
   */
  showAccountCentral: function() {
MoreLayouts.log.debug("showAccountCentral ");
    let displayDeck = MoreLayouts.e("displayDeck");
    let accountCentralBox = MoreLayouts.e("accountCentralBox");

    displayDeck.selectedPanel = accountCentralBox;

    let prefName = "mailnews.account_central_page.url";
    let acctCentralPage =
      pref.getComplexValue(prefName, MoreLayouts.Ci.nsIPrefLocalizedString).data;

    // Set the layout, if necessary.  At least print the layout message.
    MoreLayouts.gAccountCentralLoaded = true;
    MoreLayouts.updateMailPaneConfig(true);

    // Disable layout menuitems in accountcentral.
    MoreLayouts.enableLayoutMenuItems(false);

    // Frame reference lost with WideThreadPaneConfig, load it thusly..
//    window.frames["accountCentralPane"].location.href = acctCentralPage;
    accountCentralBox.firstChild.contentDocument.location.href = acctCentralPage;
  },

  /**
   * New FolderDisplayWidget._showThreadPane().
   */
  showThreadPane: function() {
MoreLayouts.log.debug("showThreadPane ");
    let displayDeck = MoreLayouts.e("displayDeck");
    let threadPaneBox = MoreLayouts.e("threadPaneBox");

    if (displayDeck.selectedPanel == threadPaneBox) {
      // It seems makeActive() fires twice merely switching from account to a
      // folder; don't need to go into updateMailPaneConfig.
      return;
    }
MoreLayouts.log.debug("showThreadPane continue ");

    displayDeck.selectedPanel = threadPaneBox;

    // Set the layout.
    MoreLayouts.gAccountCentralLoaded = false;
    MoreLayouts.updateMailPaneConfig(true);
  },

  /**
   * New FolderDisplayWidget._updateThreadDisplay().
   */
  updateThreadDisplay: function() {
MoreLayouts.log.debug("updateThreadDisplay ");
    if (this.active && this.view.dbView) {
MoreLayouts.log.debug("updateThreadDisplay: dbView ");
      UpdateSortIndicators(this.view.dbView.sortType, this.view.dbView.sortOrder);
      SetNewsFolderColumns();
    }

    // When switching tabs and showing threadpane, make sure all views are
    // correct, given we may be in a message pane toggle state.  Each folder
    // tab may have its own state.
    let tabmail = MoreLayouts.e("tabmail");
    if (tabmail && tabmail.currentTabInfo &&
        tabmail.currentTabInfo.mode.type == "folder") {
      if ("messagePaneState" in tabmail.currentTabInfo.folderDisplay)
        MoreLayouts.gCurrentMessagePaneConfig =
            tabmail.currentTabInfo.folderDisplay.messagePaneState;
      else
        MoreLayouts.gCurrentMessagePaneConfig = MoreLayouts.kMessagePaneNormal;
      MoreLayouts.MessagePaneMaxShowHide(MoreLayouts.gCurrentMessagePaneConfig);

      if (MoreLayouts.gCurrentMessagePaneConfig == MoreLayouts.kMessagePaneNormal)
        MoreLayouts.enableLayoutMenuItems(true);
    }
    else {
      MoreLayouts.enableLayoutMenuItems(false);
    }
  },

  /**
   * Set enable/disable status of View->Layout menuitems.
   * @param aEnable Bool enable/disable.
   */
  enableLayoutMenuItems: function(aEnable) {
    // Enable layout menuitems in threadpane.
    let layoutItems = document.getElementsByAttribute("name", "viewlayoutgroup");
    for (let i = 0; i < layoutItems.length; i++)
      if (aEnable)
        layoutItems[i].removeAttribute("disabled");
      else
        layoutItems[i].setAttribute("disabled", true);
  },

  /**
   * Replace UpdateMailPaneConfig().
   */
  updateMailPaneConfig: function(aMsgWindowInitialized, aConfig) {
    let paneConfig = aConfig == null ? this.paneConfig : aConfig;

this.log.debug("paneConfig:gCurrentPaneConfig:gAccountCentralLoaded " +
               paneConfig+":"+this.gCurrentPaneConfig+":"+this.gAccountCentralLoaded);

    // Set to standard/classic if in accountcentral and widethread or stacked
    // layout, so accountcentral displays correctly.
    if (this.gAccountCentralLoaded && //aMsgWindowInitialized &&
        this.dynamicViewResetDOMType.indexOf(paneConfig) != -1)
      paneConfig = kStandardPaneConfig;

    // Make sure it's valid.
    paneConfig = this.dynamicIdsMessagePane[paneConfig] ? paneConfig :
                                                          kStandardPaneConfig;

    // Set view name for css selectors.
    let viewName =
      this.gAccountCentralLoaded ? this.dynamicViewName[this.kAccountCentral] :
                                   this.dynamicViewName[paneConfig];
    this.e("mailContent").setAttribute("layout", viewName);

    if (!aMsgWindowInitialized) {
      // Some init stuff.
      this.gCurrentMessagePaneConfig = this.kMessagePaneNormal;
      this.gPriorMessagePaneConfig = this.kMessagePaneMax;
    }

    this.log.info("layout --> " + viewName);

    if (paneConfig == this.gCurrentPaneConfig)
      // No change required.
      return;

    let rerootThreadPane = false;
    let desiredIdMessagePane = this.dynamicIdsMessagePane[paneConfig];
    let desiredIdThreadPane = this.dynamicIdsThreadPane[paneConfig];
    let desiredIdThreadPaneSplitter = this.dynamicIdsThreadPaneSplitter[paneConfig];

    // threadpane+acct central.
    let displayDeck = this.e("displayDeck");
    // GetMessagePane();
    gMessagePane = this.e("messagepanebox");
    // GetMessagePaneBoxWrapper();
    gMessagePaneWrapper = this.e("messagepaneboxwrapper");
    // GetThreadAndMessagePaneSplitter();
    gThreadAndMessagePaneSplitter = this.e("threadpane-splitter");

this.log.debug("messagePaneBoxWrapper: parent:desiredParent - " +
               gMessagePaneWrapper.parentNode.id+":"+desiredIdMessagePane);

    if (gMessagePaneWrapper.parentNode.id != desiredIdMessagePane) {
      ClearAttachmentList();
      let hdrToolbox = this.e("header-view-toolbox");
      let hdrToolbar = this.e("header-view-toolbar");
      let firstPermanentChild = hdrToolbar.firstPermanentChild;
      let lastPermanentChild = hdrToolbar.lastPermanentChild;
      let desiredParent = this.e(desiredIdMessagePane);

      // "Here the message pane including the header pane is moved..."
      let cloneToolboxPalette;
      let cloneToolbarset;
      if (hdrToolbox.palette) {
        cloneToolboxPalette = hdrToolbox.palette.cloneNode(true);
      }
      if (hdrToolbox.toolbarset) {
        cloneToolbarset = hdrToolbox.toolbarset.cloneNode(true);
      }

      // "See Bug 381992. The ctor for the browser element will fire again..."
      this.e("messagepane").destroy();

      desiredParent.appendChild(gThreadAndMessagePaneSplitter);
      desiredParent.appendChild(gMessagePaneWrapper);

      hdrToolbox.palette  = cloneToolboxPalette;
      hdrToolbox.toolbarset = cloneToolbarset;
      hdrToolbar = this.e("header-view-toolbar");
      hdrToolbar.firstPermanentChild = firstPermanentChild;
      hdrToolbar.lastPermanentChild = lastPermanentChild;
    }

this.log.debug("displayDeck: parent:desiredParent - " +
               displayDeck.parentNode.id+":"+desiredIdThreadPane);
this.log.debug("threadpane-splitter: parent:desiredParent - " +
               gThreadAndMessagePaneSplitter.parentNode.id+":"+
               desiredIdThreadPaneSplitter);

    if (displayDeck.parentNode.id != desiredIdThreadPane) {
      let desiredParent = this.e(desiredIdThreadPane);
      let desiredParentSplitter = this.e(desiredIdThreadPaneSplitter);
      switch (paneConfig) {
        case kStandardPaneConfig: // 0
        case this.kWideThreadPaneConfig: // 3
          desiredParent.insertBefore(displayDeck, desiredParent.lastChild);
          break;
        case kWidePaneConfig: // 1
        case kVerticalPaneConfig: // 2
        case this.kStackedPaneConfig: // 4
          desiredParent.appendChild(displayDeck);
          break;
      }

      // Make sure splitter is in the right place.
      desiredParentSplitter.insertBefore(gThreadAndMessagePaneSplitter,
                                         desiredParentSplitter.lastChild);

      rerootThreadPane = true;
    }

    // Set element orient states for a given view.
    gThreadAndMessagePaneSplitter.orient =
      this.dynamicThreadPaneSplitterOrient[paneConfig] ? "vertical" : "horizontal";

    // Make sure dimensions always maintained as set.  Also set flex on a tab
    // change for non 3pane.
    gMessagePaneWrapper.setAttribute("flex", this.dynamicMessagePaneBoxFlex[paneConfig]);
    displayDeck.setAttribute("flex", this.dynamicThreadPaneBoxFlex[paneConfig]);
//    this.e("folderPaneBox").removeAttribute("flex");

    // Record the new configuration, set the pref in case changed manually
    // (unless we're switching layout for fullscreen).
    this.gCurrentPaneConfig = paneConfig;
    if (!this.gAccountCentralLoaded && !aConfig)
      this.paneConfig = paneConfig;

    // Reset msgWindow after move.
    if (aMsgWindowInitialized) {
      this.rerootMsgWindow();

      if (rerootThreadPane)
        // Also reset threadpane references after move, if necessary.
        this.rerootThreadPane();
    }

    if (gDBView && GetNumSelectedMessages() == 1)
      gDBView.reloadMessage();

    // "The quick filter bar gets badly lied to due...."
    setTimeout(function UpdateMailPaneConfig_deferredFixup() {
      let threadPaneBox = document.getElementById("threadPaneBox");
      let overflowNodes = threadPaneBox.querySelectorAll("[onoverflow]");
      for (let iNode = 0; iNode < overflowNodes.length; iNode++) {
        let node = overflowNodes[iNode];
        if (node.scrollWidth > node.clientWidth)
          node.onoverflow();
        else if (node.onresize)
          node.onresize();
      }
    }, 1500);

  },

  /**
   * Set splitter collapse states to effect the passed message pane state parm.
   * @param aState, string in "normal", "min", "max".
   */
  MessagePaneMaxShowHide: function(aState) {
this.log.debug("MessagePaneMaxShowHide: mode:current:prior:setState - " +
               this.messagePaneToggleMode+":"+
               this.gCurrentMessagePaneConfig+":"+
               this.gPriorMessagePaneConfig+":"+aState);

    let flex;
    let widethread = this.gCurrentPaneConfig == this.kWideThreadPaneConfig;
    let stacked = this.gCurrentPaneConfig == this.kStackedPaneConfig;
    let threadSplitter = this.e("threadpane-splitter");
    let folderSplitter = this.e("folderpane_splitter");
    let messagepanebox = this.e("messagepanebox");
    let messagepaneboxwrapper = this.e("messagepaneboxwrapper");
    let folderPaneBox = this.e("folderPaneBox");

    if (aState == this.kMessagePaneNormal) {
      if (widethread || stacked)
        folderPaneBox.removeAttribute("flex");

      this.e("threadContentArea").removeAttribute("collapsed");
      threadSplitter.setAttribute("collapse", "after");
      threadSplitter.setAttribute("state", "open");

      folderSplitter.setAttribute("collapse", "before");
      folderSplitter.setAttribute("state", "open");

      let flex = this.dynamicMessagePaneBoxFlex[this.gCurrentPaneConfig];
      messagepaneboxwrapper.setAttribute("flex", flex);
      this.enableLayoutMenuItems(true);
    }
    else if (aState == this.kMessagePaneMin) {
      if (widethread || stacked) {
        threadSplitter.setAttribute("state", "open");

        folderSplitter.setAttribute("collapse", "after");
        folderSplitter.setAttribute("state", "collapsed");
        folderPaneBox.setAttribute("flex", 1);
      }
      else {
        folderSplitter.setAttribute("state", "open");
        threadSplitter.setAttribute("state", "collapsed");
      }

      this.enableLayoutMenuItems(false);
    }
    else if (aState == this.kMessagePaneMax) {
      if (GetNumSelectedMessages() < 1)
        return false;

      threadSplitter.removeAttribute("substate");
      threadSplitter.setAttribute("collapse", "before");
      threadSplitter.setAttribute("state", "collapsed");

      folderSplitter.removeAttribute("substate");
      folderSplitter.setAttribute("collapse", "before");
      folderSplitter.setAttribute("state", "collapsed");

      flex = MoreLayouts.dynamicMessagePaneBoxFlex[kStandardPaneConfig];
      messagepaneboxwrapper.setAttribute("flex", flex);
      this.enableLayoutMenuItems(false);
    }

    messagepanebox.setAttribute("togglestate", aState);
    return true;
  },

  /**
   * An F8 three way max-normal-min message pane toggle function.
   */
  MsgToggleMessagePane: function() {
    let setState;
    let tabmail = this.e("tabmail");
    let currentTabFolderDisplay = null;
    let currentMessagePaneConfig = this.gCurrentMessagePaneConfig;
    let priorMessagePaneConfig = this.gPriorMessagePaneConfig;

    if (tabmail && tabmail.currentTabInfo &&
        tabmail.currentTabInfo.folderDisplay) {
      currentTabFolderDisplay = tabmail.currentTabInfo.folderDisplay;
      if (currentTabFolderDisplay.messagePaneStatePrior)
        priorMessagePaneConfig = currentTabFolderDisplay.messagePaneStatePrior;
    }

    if (currentMessagePaneConfig == this.kMessagePaneMin ||
        currentMessagePaneConfig == this.kMessagePaneMax)
      setState = this.kMessagePaneNormal;
    else if (this.messagePaneToggleMode == this.kToggleCycle)
      setState = priorMessagePaneConfig == this.kMessagePaneMax ?
                 this.kMessagePaneMin : this.kMessagePaneMax;
    else
      setState = this.messagePaneToggleMode == this.kToggleNormalMin ?
                 this.kMessagePaneMin : this.kMessagePaneMax;

    if (!this.MessagePaneMaxShowHide(setState))
      return;

    this.gPriorMessagePaneConfig = this.gCurrentMessagePaneConfig;
    this.gCurrentMessagePaneConfig = setState;
    if (currentTabFolderDisplay) {
      currentTabFolderDisplay.messagePaneStatePrior =
          currentTabFolderDisplay.messagePaneState;
      currentTabFolderDisplay.messagePaneState = setState;
    }

    ChangeMessagePaneVisibility(IsMessagePaneCollapsed());
    SetFocusThreadPaneIfNotOnMessagePane();
  },

  /**
   * An F11 fullscreen function.
   */
  BrowserFullScreen: function() {
this.log.debug("BrowserFullScreen: "+window.fullScreen);
    let fullscreen = window.fullScreen;
    let desiredPaneConfig;
    let fullscreenPaneConfig = this.fullscreenPaneConfig;
    if (!fullscreen && fullscreenPaneConfig != -1 &&
        this.dynamicIdsMessagePane[fullscreenPaneConfig])
      desiredPaneConfig = fullscreenPaneConfig;
    else
      desiredPaneConfig = this.paneConfig;
    if (desiredPaneConfig != this.gCurrentPaneConfig) {
      this.updateMailPaneConfig(true, desiredPaneConfig);
this.log.debug("BrowserFullScreen: done updateMailPaneConfig - "+
               this.gCurrentPaneConfig);
    }

    this.e("mail-toolbox").setAttribute("collapsed", !fullscreen);
    this.e("status-bar").setAttribute("collapsed", !fullscreen);

    // If not in fullscreen, set to fullscreen, and vice versa.
    window.fullScreen = !fullScreen;
    this.MessagePaneMaxShowHide(this.gCurrentMessagePaneConfig);
  }
};

/* Override Tb native function */
function UpdateMailPaneConfig(aMsgWindowInitialized) {
  MoreLayouts.updateMailPaneConfig(aMsgWindowInitialized);
}

/* Override Tb native function */
function MsgToggleMessagePane() {
  let currentTabInfo = document.getElementById("tabmail").currentTabInfo;
  if (currentTabInfo.mode.name != "folder")
    // Bail without doing anything if we are not a folder tab.
    return;

  MoreLayouts.MsgToggleMessagePane();
}

/* Override Tb native function */
FolderDisplayWidget.prototype._showAccountCentral = MoreLayouts.showAccountCentral;

/* Override Tb native function */
FolderDisplayWidget.prototype._showThreadPane = MoreLayouts.showThreadPane;

/* Override Tb native function */
FolderDisplayWidget.prototype._updateThreadDisplay = MoreLayouts.updateThreadDisplay;

function IsMessagePaneCollapsed() {
  if (MoreLayouts.gCurrentMessagePaneConfig == MoreLayouts.kMessagePaneMax)
    return false;
  if (MoreLayouts.gCurrentPaneConfig == MoreLayouts.kWideThreadPaneConfig ||
      MoreLayouts.gCurrentPaneConfig == MoreLayouts.kStackedPaneConfig)
    return document.getElementById("folderpane_splitter")
                   .getAttribute("state") == "collapsed";
  return document.getElementById("threadpane-splitter")
                 .getAttribute("state") == "collapsed";
}

/* Override Tb native function */
var view_initTb = view_init;
view_init = function view_init() {
  view_initTb();

  document.getElementById("menu_fullScreen_ML").
           setAttribute("checked", window.fullScreen);
}
